/*
 * is_ws.h
 *
 * Copyright (c) Chris Putnam 2003-2018
 *
 * Source code released under the GPL version 2
 *
 */
#ifndef IS_WS_H
#define IS_WS_H

extern int is_ws( char ch );
extern char *skip_ws( char *p );
extern char *skip_notws( char *p );

#endif

